/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* main.cpp
   Gabriel M. Beddingfield
   2008-02-10

   Driver to test different tap tempo models.

 */

#include <iostream>
#include <fstream>
#include <iomanip>
#include <deque>
#include <cmath>
#include <ctime>
#include <sys/time.h>

#include "../../src/TapTempo/TapTempoInterface.hpp"
#include "../../src/TapTempo/TapTempoAverage.hpp"
#include "../../src/TapTempo/TapTempoLeastSquares.hpp"
#include "../../src/TapTempo/TapTempoMedian.hpp"
#include "../../src/TapTempo/TapTempoAvgTempo.hpp"


using namespace std;
using namespace InConcert;

#warning "The usecs_t typedef should match the one in collect_data.cpp"
typedef unsigned long long usecs_t;  // microseconds

int main(int argc, char* argv[])
{
    typedef TapTempoInterface <usecs_t, double> tap_tempo_t;
    typedef deque<tap_tempo_t*> tappers_t;
    tappers_t tappers;
    deque<usecs_t> next_beat;
    deque<usecs_t> cycles_per_beat;
    deque<double> diffs;
    deque<double> diffsqrd;
    ifstream ifile;

    if (argc != 2) {
        cout << endl
             << "usage: " << argv[0] << " data_file" << endl
             << endl;
        exit(0);
    } else {
        ifile.open(argv[1]);
        if (ifile.fail()) {
            cout << "could not open file " << argv[1] << "for reading" << endl
                 << "...aborting." << endl
                 << endl;
            exit(0);
        }
    }

    // Set up the test structures.
    tappers.push_back(new TapTempoAverage<usecs_t, double>);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);
    tappers.push_back(new TapTempoLeastSquares<usecs_t, double>);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);
    tappers.push_back(new TapTempoMedian<usecs_t, double>);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);
    tappers.push_back(new TapTempoAvgTempo<usecs_t, double>);
    next_beat.push_back(0);
    cycles_per_beat.push_back(0);
    diffs.push_back(0.0);
    diffsqrd.push_back(0.0);

    cout << "Tap Tempo Experiments (postprocessing)" << endl
	 << "2008-02-10 Gabriel M. Beddingfield" << endl
	 << endl;

    int getting_taps;
    string speed;
    usecs_t now;
    int j;
    bool not_done = true;
    while(ifile.good()) {

        ifile >> getting_taps;
        if (ifile.eof()) {
            break;
        }

	for(j=0; j<tappers.size(); ++j) {
	    tappers[j]->setTapsRequired(getting_taps-1);
	}

        ifile >> speed;
        if (speed[speed.length()-1] == ':') {
            speed.erase(speed.length()-1, 1);
        }

	cout << "Please give " << getting_taps << " taps "
	     << speed << endl;

	while(getting_taps > 0) {
            ifile >> now;
	    cout << "Tap @ " << dec << now << endl;

	    // Register taps.
	    --getting_taps;
            if (getting_taps == 0) continue;
	    for(j=0; j<tappers.size(); ++j) {
		next_beat[j] = tappers[j]->tap(now);
                if(next_beat[j] > 0) cycles_per_beat[j] = tappers[j]->getTempoInterval();
	    }
	}

	for(j=0; j<tappers.size(); ++j) {
            double diff;
            diff = (double(now)-double(next_beat[j]))/1000.0;
	    cout << tappers[j]->name() << ": "
		 << "cpb = " << dec << cycles_per_beat[j] << " "
		 << "next = " << dec << next_beat[j] << " "
		 << "diff = " << diff << " ms"
		 << endl;
            diffs[j] += fabs(diff);
            diffsqrd[j] += diff*diff;
	    next_beat[j] = 0;
	    cycles_per_beat[j] = 0;
	}
    }

    ifile.close();

    cout << "Final stats:" << endl;
    for(j=0; j<tappers.size(); ++j) {
        cout << tappers[j]->name() << ": "
             << diffs[j] << " ms / " 
             << diffsqrd[j] << " ms^2"
             << endl;
    }

    while(tappers.size() > 0) {
	delete tappers.front();
	tappers.pop_front();
    }

    return 0;
}
