#ifndef __TAPTEMPOAVGTEMPO_HPP__              /*-*- indent-tabs-mode:nil;-*-*/
#define __TAPTEMPOAVGTEMPO_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempoAvgTempo.hpp
   Gabriel M. Beddingfield
   2008-08-16

   This model is based on the one in the development version
   of Hydrogen 0.9.4 (pre).  I'm not sure who originally came
   up with it.  I've added the Tn+1 logic... that model is
   acutally just a rolling average bpm -- so there is no
   sync logic (Tn+1).

   Ij = Tj-T(j-1)
   Rj = 1/Ij
   Tavg = n/sum(Ri)
   Tn+1 = Tn + Tavg

 */

#include <QMutex>
#include <QMutexLocker>
#include <cassert>
#include "TapTempoBaseImpl.hpp"

namespace InConcert
{
    template <typename ScalarT, typename TempoT = double>
    class TapTempoAvgTempo : public TapTempoBaseImpl<ScalarT, TempoT>
    {
        ScalarT m_last_tap;
        TempoT m_sum_rates;
        unsigned m_taps_received;
        QMutex m_mutex;
        typedef TapTempoBaseImpl<ScalarT, TempoT> base_class;

    public:
        TapTempoAvgTempo() :
            base_class("TapTempoAvgTempo"),
            m_last_tap(0),
            m_sum_rates(0),
            m_taps_received(0)
            { }
        virtual ~TapTempoAvgTempo() { }
        virtual void invalidate(void) {
            QMutexLocker mx(&m_mutex);
            m_last_tap = 0;
            m_sum_rates = 0;
            m_taps_received = 0;
            base_class::invalidate();
        }

        virtual ScalarT tap(ScalarT tap) {
            if (this->reset() || (m_taps_received == this->getTapsRequired())) {
                invalidate();
            } else if (this->getTimeout()
                       && (tap - m_last_tap) >= this->getTimeout()) {
                invalidate();
            }
            QMutexLocker mx(&m_mutex);

            ScalarT rv = 0, interval = 0;
            TempoT rate;

            if (m_taps_received == 0) {
                m_last_tap = tap;
                ++m_taps_received;
                return rv;
            }
            interval = tap - m_last_tap;
            rate = 1.0 / interval;
            m_sum_rates += rate;
            m_last_tap = tap;
            ++m_taps_received;
            if(m_taps_received == this->getTapsRequired()) {
                TempoT inv_avg_rate = (TempoT)(m_taps_received - 1) / m_sum_rates;
                assert( inv_avg_rate > 0 );
                this->setTempoByInterval(inv_avg_rate);
                rv = tap + inv_avg_rate;
                this->queueReset();
            }
            return rv;
        }

    }; // class TapTempoAverage

} // namespace InConcert    
#endif // __TAPTEMPOAVGTEMPO_HPP__
