#ifndef __TAPTEMPOLEASTSQUARES_HPP__
#define __TAPTEMPOLEASTSQUARES_HPP__
/**********************************************-*- indent-tabs-mode:nil; -*-
 *                                                                         *
 *   InConcert - Tap Tempo Synchronization for Live Perforances            *
 *                                                                         *
 *   Copyright (C) 2008 by Gabriel M. Beddingfield                         *
 *   gabriel@teuton.org                                                    *
 *                                                                         *
 *   "For of him [God], and through him, and unto him, are all things.     *
 *   To him be the glory for ever. Amen." (Romans 11:36)                   *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; version 2 of the License, or any later  *
 *   version.                                                              *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/* TapTempoLeastSquares.hpp
   Gabriel M. Beddingfield
   2008-08-16

*/

#include <cassert>
#include "TapTempoBaseImpl.hpp"
#include <QMutex>

namespace InConcert
{
    // TapTempoLeastSqures
    // Tn+1 determined by a least squares fit of T0..Tn
    template <typename ScalarT, typename TempoT = double>
    class TapTempoLeastSquares : public TapTempoBaseImpl<ScalarT, TempoT>
    {
	unsigned m_taps_received;
	ScalarT m_last_tap;  // In seconds (frame/m_frame_rate)
	ScalarT m_sum_taps;
	ScalarT m_sum_i_taps;
	QMutex m_mutex;
	typedef TapTempoBaseImpl<ScalarT, TempoT> base_class;

    public:
	TapTempoLeastSquares() :
	    base_class("TapTempoLeastSquares"),
	    m_taps_received(0),
	    m_last_tap(0),
	    m_sum_taps(0),
	    m_sum_i_taps(0)
	    { }

	virtual void invalidate(void) {
	    QMutexLocker mx(&m_mutex);
	    m_sum_taps = m_sum_i_taps = 0.0;
	    m_taps_received = 0;
	    base_class::invalidate();
	}

	ScalarT tap(ScalarT tap) {
	    if (this->reset() || m_taps_received == this->getTapsRequired()) {
		invalidate();
	    } else if (this->getTimeout()
		       && (tap - m_last_tap) >= this->getTimeout()) {
		invalidate();
	    }
	    QMutexLocker mx(&m_mutex);

	    assert(tap > m_last_tap);
    
	    // Push the tap onto the accumulators
	    // Asses ++m_taps_received because of zero-offset indices.
	    // See derivation.
	    m_last_tap = tap;
	    m_sum_taps += tap;
	    m_sum_i_taps += (ScalarT)m_taps_received * tap;
	    ++m_taps_received;
    
	    // Check if we've got the last one.
	    if(m_taps_received == this->getTapsRequired()) {
		this->queueReset();
		return lastTap();
	    }
	    return 0;
	}

    private:
    
	// lastTap()
	//
	// Predicts tap and rate based on previous N taps using a least
	// squares analysis for each tap.  Assumes that an ideal tempo is
	// desired, but represented by imperfect taps that follow a normal
	// distribution on the error.
	//
	// Presume N+1 ideal tempo B0, B1, B2, ... BN being represented by N+1
	// imperfect taps T0, T1, T2, ... TN.  Each Bi is evenly spaced, such
	// that:
	//
	// Bi = B0 + (i*D)
	//
	// Where D is the space between beats.  Thus, we want to find the B0
	// and D that best represents that taps that we've been given.
	//
	// Since each Ti represents each Bi, it has an error associated with
	// it.  Thus Bi = Ti + Ei, or:
	//
	// Ei = Bi - Ti
	//
	// Where Ei is the error of the tap relative to the ideal tempo.  Without
	// going into great detail, Ei^2 is directly related to the statistical
	// probability distribution (bell curve, standard devation).
	//
	// We get an indication of the total error by summing the squares:
	//
	// SUM(Ei^2) = E0^2 + E1^2 + E2^2 + ... + EN^2
	//
	// Which, when substituting the above equations, becomes:
	//
	// SUM(Ei^2) = (B0-T0)^2 + (B1-T1)^2 + ... + (BN-TN)^2
	//           = (B0-T0)^2 + (B0+D-T1)^2 + ... + (B0+N*D-TN)^2
	//           = SUM( (B0 + i*D - Ti)^2 ) for i=0..N
	//           = EE
	//
	// We seek to minimize this function.  This may be done by taking
	// the partial derivative of each parameter, and setting it equal to
	// zero.
	//
	// d(EE) = d(EE)/d(B0) + d(EE)/d(D)
	//
	// Where d() is the differential operator.  To minimize, we set
	// both terms equal to zero:
	//
	// d(EE)/d(B0) = 0
	//
	// d(EE)/d(D) = 0
	//
	// And solve.
	//
	// d(EE)/d(B0) = SUM( 2*(B0+i*D-Ti) ) for i=0..N
	//             = 2*SUM(B0+i*D-Ti) for i=0..N
	//             = 2*SUM(B0) + 2*SUM(i*D) - 2*SUM(Ti), all for i=0..N
	//             = 2*B0*(N+1) + 2*D*SUM(i) - 2*SUM(Ti), all for i=0..N
	// 
	// d(EE)/d(D)  = SUM( 2*(B0+i*D-Ti)*i ) for i=0..N
	//             = SUM( 2*i*B0 + 2*i^2*D - 2*i*Ti ) for i=0..N
	//             = SUM( 2*i*B0 ) + SUM(2*i^2*D) - SUM(2*i*Ti), all for i=0..N
	//             = 2*B0*SUM(i) + 2*D*SUM(i^2) - 2*SUM(i*Ti), all for i=0..N
	//
	// Setting these to zero gives us a linear system of equation.  Note that
	// from here on out, SUM(...) is implied to be over the range i=0..N.
	//
	// 0 = 2*(N+1)  * B0 + 2*SUM(i)   * D - 2*SUM(Ti)      {d(EE)/d(B0)}
	// 0 = 2*SUM(i) * B0 + 2*SUM(i^2) * D - 2*SUM(i*Ti)    {d(EE)/d(D)}
	//
	// [ 0   = [ 2*(N+1)    2*SUM(i)     * [ B0     - 2 * [ SUM(Ti)
	//   0 ]     2*SUM(i)   2*SUM(i^2) ]     D  ]           SUM(i*Ti) ]
	//
	// Multiplying all bey 1/2 and moving the constant term to the left:
	//
	// [ SUM(Ti)     = [ (N+1)    SUM(i)     * [ B0
	//   SUM(i*Ti) ]     SUM(i)   SUM(i^2) ]     D  ]
	//
	// For simplicity, let the coefficient matrix be represented by:
	//
	// [ m1  m2
	//   m2  m4 ]  (note that m3 = m2, so it's omitted)
	//
	// Note that it's simple to prove mi > 0 for all N > 0.
	//
	// Solving by row reduction, we eliminate the bottom left entry by
	// multiplying the first row by -m2/m1 and adding it to the second
	// row.
	//
	// [ SUM(Ti)                    = [ m1  m2             [ B0
	//   SUM(i*Ti)-(m2/m1)SUM(Ti) ]      0  m4-(m2^2/m1) ]    D ]
	// 
	// Solving by back-substitution:
	//
	// D = [SUM(i*Ti)-(m2/m1)SUM(Ti)] / [m4-(m2^2/m1)]
	//
	// SUM(Ti) = m1*B0 + m2*D
	// m1*B0 = SUM(Ti) - m2*D
	// B0 = (SUM(Ti) - m2*D)/m1 = (1/m1)*SUM(Ti) - (m2/m1)*D
	//
	// Notice that B0 and D only depend on the following 3 terms:
	//
	// Ca = (m2/m1)
	// Cb = 1/(m4-(m2^2/m1))
	//    = m1 / (m4*m1 - m2^2)
	// Cc = (1/m1)
	//
	// Now, SUM(i) and SUM(i^2) (and thus, each mi) can be calculated
	// knowing only N.  I once derived the solution for SUM(i), but I
	// verified and found the solution in an online database called "The
	// On-Line Encyclopedia of Integer Sequences"
	// http://www.research.att.com/~njas/sequences
	//
	// There, I found:
	//
	// SUM(i) for i=0..N = N*(N+1)/2
	// SUM(i^2) for i=0..N = N*(N+1)*(2N+1)/6
	//
	// Reducing Ca, Cb, and Cc...
	//
	// m1 = (N+1)
	// m2 = SUM(i) = N*(N+1)/2
	// m4 = SUM(i^2) = N*(N+1)*(2N+1)/6
	// Ca = (m2/m1)
	//    = (N*(N+1)/2) / (N+1)
	//    = N/2
	// Cb = 1/(m4-(m2^2/m1))
	//    = m1 / (m4*m1 - m2^2)
	//    = (N+1) / [(N*(N+1)*(2N+1)/6) * (N+1) - (N*(N+1)/2)^2]
	//    = (N+1) / [ N*(N+1)^2*(2N+1)/6 - N^2*(N+1)^2/4 ]
	//    =  1 / [N*(N+1)*(2N+1)/6 - N^2*(N+1)/4]
	//    =  1 / [N*(N+1)*(2N+1)*4/24 - N^2*(N+1)*6/24]
	//    = 24 / [N*(N+1)*(2N+1)*4 - N^2*(N+1)*6]
	//    = 24 / N*(N+1) * [ 4*(2N+1) - 6*N ]
	// Cc = (1/m1)
	//    = 1/(N+1)
	//
	// 
	// D = [SUM(i*Ti)-(m2/m1)SUM(Ti)] / [m4-(m2^2/m1)]
	//   = [SUM(i*Ti)-Ca*SUM(Ti)] * Cb
	// B0 = Cc*SUM(Ti) - Ca*D
	//
	// To reduce calculation and lag time, Ca, Cb, and Cc can be
	// calculated before-hand.  Meanwhile, SUM(Ti) and SUM(i*Ti) can be
	// tabulated as the taps come in.
	//
	ScalarT lastTap(void)
	{
	    double D, B0;
	    double next_frame;
	    
	    if(this->getTapsRequired() == 2) {
		B0 = m_sum_taps - m_last_tap;
		D = m_last_tap - B0;
		next_frame = B0 + 2.0*D;
	    } else {
		double Ca, Cb, Cc;
		double N = (double)(this->getTapsRequired()) - 1.0;
		double Np1 = (double)(this->getTapsRequired()); // N+1
		
		Ca = N/2.0;
		Cb = N*(N+1.0) * ( 4.0*(2.0*N+1) - (6.0)*N );
		Cb = 24.0 / Cb;
		Cc = 1/Np1;
		
		D = (m_sum_i_taps - Ca * m_sum_taps) * Cb;
		B0 = Cc*m_sum_taps - Ca*D;
		
		assert(D > .001);  // Crazy tapping.
		next_frame = B0 + Np1*D;
	    }
	    setTempoByInterval(ScalarT(D));
	    return ScalarT(next_frame);
	}
    };

} // namespace InConcert

#endif // __TAPTEMPOLEASTSQUARES_HPP__
